
// double to float conv
#pragma warning(disable:4244)

#ifdef _DEBUG
#include <assert.h>
#endif

#define HVCtrX 0.0
#define HVCtrY 0.0
Float HoleFoc=2.0f;
#define HoleFocX HoleFoc
#define HoleFocY HoleFoc

#define SHFTHTPS 8
#define HOLEGEN 25
#define HoleNbParImg 64
#define HoleNbImg 64
#define HOLEVIT 8

#define TUNNELCOLORRND 0.15f
#define TUNNELCOLORFACT 0.9999f

inline int rnd(int n)
{
	return rand()*n/(RAND_MAX+1);
}

inline float rnd(float n)
{
	return (float)rand()*(1.0f/(RAND_MAX+1))*n;
}

struct THole
{
	Float u,v; // point position on the plane
	float c1,c2; // color coefficients
};
THole Hole[HoleNbImg][HoleNbParImg];
THole RefHole[HoleNbParImg];

inline Float operator*(const SVector3D &a,const SVector3D &b)
{
	return a.DotProduct(b);
}
inline SVector3D operator*(const SVector3D &a,Float b)
{
	return SVector3D(a.x*b,a.y*b,a.z*b);
}
inline SVector3D &operator*=(SVector3D &a,Float b)
{
	a.Scale(b);
	return a;
}
/*inline SVector3D operator+(const SVector3D &a,const SVector3D &b)
{
	return SVector3D(a.x+b.x,a.y+b.y,a.z+b.z);
}*/

class SMatrix3D
{
public:
	SVector3D x,y,z;
	SMatrix3D() :x(1,0,0),y(0,1,0),z(0,0,1) {}
	void FromAngle(float ax,float ay,float az)
	{
		Float cx=cos(ax); Float sx=sin(ax);
		Float cy=cos(ay); Float sy=sin(ay);
		Float cz=cos(az); Float sz=sin(az);
		// 3D rotation matrix
		x.x=cy*cz; x.y=cx*sz*cy+sx*sy; x.z=sx*sz*cy-cx*sy;
		y.x=-sz;	 y.y=cx*cz;	         y.z=sx*cz;
		z.x=cz*sy; z.y=cx*sz*sy-sx*cy; z.z=sx*sz*sy+cx*cy;
	}
	void Apply(const SVector3D &v,SVector3D *out)
	{
		out->x=v*x;
		out->y=v*y;
		out->z=v*z;
	}
	void RApply(SVector3D const &v,SVector3D *out)
	{
		out->x=v.x*x.x+v.y*y.x+v.z*z.x;
		out->y=v.x*x.y+v.y*y.y+v.z*z.y;
		out->z=v.x*x.z+v.y*y.z+v.z*z.z;
	}
	void Apply(SVector3D *v)
	{
		SVector3D i(*v);
		Apply(i,v);
	}
	void RApply(SVector3D *v)
	{
		SVector3D i(*v);
		RApply(i,v);
	}
};

struct THoleTraj
{
	SVector3D a; // angles
	SVector3D o,m,n; // base vectors of the plane
	Float s; // size of the circle
};
THoleTraj HoleTraj[HoleNbImg];

void HoleInitPlan(int p,int t,float ss=1.0)
{
	float c1,c2;
	float s=ss;

	// calc color and position
	for(int i=0;i<HoleNbParImg;i++)
	{
			c1=RefHole[i].c1+rnd(2*TUNNELCOLORRND)-TUNNELCOLORRND;
			if (c1<0) c1=0; else if (c1>1) c1=1;
			c2=RefHole[i].c2+rnd(2*TUNNELCOLORRND)-TUNNELCOLORRND;
			if (c2<0) c2=0; else if (c2>1) c2=1;

			Hole[p][i].c1=c1;
			Hole[p][i].c2=c2;
			s=ss*(2.0f-c1);
			Hole[p][i].u=RefHole[i].u*s;
			Hole[p][i].v=RefHole[i].v*s;
	}

	// color smoothing
	for(i=0;i<HoleNbParImg;i++)
	{
			RefHole[i].c1=(Hole[p][i].c1+Hole[p][(i+1)&(HoleNbParImg-1)].c1
			               +Hole[p][(i-1)&(HoleNbParImg-1)].c1)*TUNNELCOLORFACT/3;
			RefHole[i].c2=(Hole[p][i].c2+Hole[p][(i+1)&(HoleNbParImg-1)].c2
			               +Hole[p][(i-1)&(HoleNbParImg-1)].c2)*TUNNELCOLORFACT/3;
	}

	HoleTraj[p].s=s*HOLEGEN;

	// calc trajectory (based on c2)

	static float c2a_0=0;
	static float c2b_0=0;

#define C2_VIT 0.01f

	float c2a=RefHole[0].c2;
	float c2b=RefHole[HoleNbParImg/2].c2;

	     if (c2a<c2a_0-C2_VIT) c2a_0-=C2_VIT;
	else if (c2a>c2a_0+C2_VIT) c2a_0+=C2_VIT;
	else c2a_0=c2a;

	     if (c2b<c2b_0-C2_VIT) c2b_0-=C2_VIT;
	else if (c2b>c2b_0+C2_VIT) c2b_0+=C2_VIT;
	else c2b_0=c2a;

	float az=2*c2a_0-1.0f; az=az*2;
	float ax=2*c2b_0-1.0f; ax=ax*2;

	HoleTraj[p].a.y=0;                            //(float)(sin(t*PI/90)*4*HOLEVIT*PI/1500);
	HoleTraj[p].a.z=(az*az*az*2)*HOLEVIT*PI/2500; //(float)(sin(t*PI/170)*2*HOLEVIT*PI/2500);
	HoleTraj[p].a.x=(ax*ax*ax*2)*HOLEVIT*PI/2500; //(float)(sin(t*PI/170)*2*HOLEVIT*PI/2500);
}
int HoleLastP; // last calculed plane
int HoleNbImgA; // nb of rendered planes
Bool SinHole=False;
Bool StopHole=False;
int HoledP;

void HoleInit()
{
	int i;
	for (i=0;i<HoleNbParImg;i++)
	{
		RefHole[i].u=(float)(sin(i*(2*PI/HoleNbParImg))*HOLEGEN);
		RefHole[i].v=(float)(cos(i*(2*PI/HoleNbParImg))*HOLEGEN);
		RefHole[i].c1=0;
		RefHole[i].c2=0;
	}
	HoleLastP=-1;
	HoledP=1;
}

struct BBox2D
{
	Float u0,v0;
	Float u1,v1;
};
/*
 calc intersection of 2 circles
 (see bottom)
*/
void InterLnCircle(double u,double v,double w,Float *x1,Float *y1,Float *x2,Float *y2)
{
	double d=1.0/sqrt(u*u+v*v);
	u*=d; v*=d; w*=d;
#ifdef _DEBUG
	assert(abs(w)<1.0);
#endif
	d=sqrt(1.0-w*w);
	*x1=w*u-d*v; *y1=w*v+d*u;
	*x2=w*u+d*v; *y2=w*v-d*u;
#ifdef _DEBUG
	assert(abs((*x1)*u+(*y1)*v-w)<1e-6);
	assert(abs((*x2)*u+(*y2)*v-w)<1e-6);
	assert(abs(sqr(*x1)+sqr(*y1)-1.0)<1e-6);
	assert(abs(sqr(*x2)+sqr(*y2)-1.0)<1e-6);
#endif
}

void CalcBBoxPlan(int p,BBox2D *b)
{
	SVector3D o,m,n;
	o=HoleTraj[p].o;
	m=HoleTraj[p].m*HoleTraj[p].s;
	n=HoleTraj[p].n*HoleTraj[p].s;
	Float ca1,ca2,sa1,sa2;
	Float f1,f2;
	//ca(mxoy-myox)+sa(nyox-nxoy)=mynx-mxny
	InterLnCircle(o.y*m.x-o.x*m.y
	             ,o.x*n.y-o.y*n.x
							 ,m.y*n.x-m.x*n.y
							 ,&ca1,&sa1,&ca2,&sa2);
	f1=HVCtrX+HoleFocX*(o.x+m.x*sa1+n.x*ca1)/(o.y+m.y*sa1+n.y*ca1);
	f2=HVCtrX+HoleFocX*(o.x+m.x*sa2+n.x*ca2)/(o.y+m.y*sa2+n.y*ca2);
	if (f1<f2) { b->u0=f1; b->u1=f2; }
	else       { b->u0=f2; b->u1=f1; }
	//ca(mzoy-myoz)+sa(nyoz-nzoy)=mynz-mzny
	InterLnCircle(o.y*m.z-o.z*m.y
	             ,o.z*n.y-o.y*n.z
							 ,m.y*n.z-m.z*n.y
							 ,&ca1,&sa1,&ca2,&sa2);
	f1=HVCtrY-HoleFocY*(o.z+m.z*sa1+n.z*ca1)/(o.y+m.y*sa1+n.y*ca1);
	f2=HVCtrY-HoleFocY*(o.z+m.z*sa2+n.z*ca2)/(o.y+m.y*sa2+n.y*ca2);
	if (f1<f2) { b->v0=f1; b->v1=f2; }
	else       { b->v0=f2; b->v1=f1; }
}
void InterBBox(BBox2D *a,BBox2D const *b)
{
	a->u0=max(a->u0,b->u0);
	a->u1=min(a->u1,b->u1);
	a->v0=max(a->v0,b->v0);
	a->v1=min(a->v1,b->v1);
}
Bool BBoxEmpty(BBox2D const *b)
{
	return (b->u0>b->u1 || b->v0>b->v1);
}
void MkBBoxAll(BBox2D *b)
{
	b->u0=-1e10; b->u1=1e10;
	b->v0=-1e10; b->v1=1e10;
}

BBox2D BBPlan[HoleNbImg];

struct HPT
{
	float ex,ey;
	float u,v;
	float c1;
} Pt[HoleNbImg][HoleNbParImg+1];
float PtDist[HoleNbImg];

void CalcHole(int T)
{
	float ft=(T&((1<<SHFTHTPS)-1))*(1.0/(1<<SHFTHTPS));
	int it=T>>SHFTHTPS;
	int i,p,s;
	// 1st : calc planes position
	SVector3D o(0,0,0);
	SVector3D m(1,0,0);
	SVector3D v(0,HOLEVIT,0);
	SVector3D n(0,0,1);
	SVector3D a;
	SMatrix3D mm;
	BBox2D bplan,bhole;
	MkBBoxAll(&bhole);
	for (i=0;i<HoleNbImg && !BBoxEmpty(&bhole);i++)
	{
		p=(i+it)&(HoleNbImg-1);
		if (i+it>HoleLastP)
		{ // this plane need to be calculated
			if (StopHole) break; // stop plane calculation
			HoleInitPlan(p,i+it,(SinHole?(sin((i+it)*PI/10)+4.0)/4:1.0));
			HoleLastP=i+it;
		}
		if (!i)
			o+=v*(1-ft);
		else
			o+=v;
		HoleTraj[p].o=o;
		HoleTraj[p].m=m;
		HoleTraj[p].n=n;
		if (!i)
			mm.FromAngle(HoleTraj[p].a.x*(1-ft),HoleTraj[p].a.y*(1-ft),HoleTraj[p].a.z*(1-ft));
		else
			mm.FromAngle(HoleTraj[p].a.x,HoleTraj[p].a.y,HoleTraj[p].a.z);
		mm.Apply(&m);
		mm.Apply(&v);
		mm.Apply(&n);
		CalcBBoxPlan(p,&bplan); BBPlan[i]=bhole;
		InterBBox(&bhole,&bplan);
	}
	HoleNbImgA=i; if (!HoleNbImgA) return;

	// 2nd : calc points positions
	// Equations:
	// point at (u,v) on plane (O,m,n)
	// -> s>=o>+u*m>+v*n>
	// px=ox+u*mx+v*nx
	// py=oy+u*my+v*ny
	// pz=oz+u*mz+v*nz

	SVector3D pp;
	float txtv;
	float pu,pv;
	for (i=0;i<HoleNbImgA;++i)
	{
		PtDist[i]=i+1.0-ft;
		p=(i+it)&(HoleNbImg-1);
		o=HoleTraj[p].o;
		m=HoleTraj[p].m;
		n=HoleTraj[p].n;
		txtv=(i+it)*(2.0/HoleNbImg);
		for (s=0;s<=HoleNbParImg;s+=HoledP)
		{
			pu=Hole[p][s&(HoleNbParImg-1)].u;
			pv=Hole[p][s&(HoleNbParImg-1)].v;
			//pp=o+m*pu+n*pv;
			pp.x=o.x+pu*m.x+pv*n.x;
			pp.y=o.y+pu*m.y+pv*n.y;
			pp.z=o.z+pu*m.z+pv*n.z;
			if (pp.y<=0) pp.y=0.1f; // no divide by zero
			Pt[i][s].ex=HVCtrX+HoleFocX*pp.x/pp.y;
			Pt[i][s].ey=HVCtrY-HoleFocY*pp.z/pp.y;
			Pt[i][s].c1=0.25f+0.75f*Hole[p][s&(HoleNbParImg-1)].c1;
			Pt[i][s].u=s*(1.0/HoleNbParImg);
			Pt[i][s].v=txtv;
		}
	}
}

/*

!!! in french, sorry...

Optimisation de l'affichage du tunnel:
le tunnel est constitu de cercles.
on peut arreter l'affichage du  tunnel quand le reste cache tout l'cran
si on garde une zone de visibilit qui montre le zone encore visible de l'cran
aprs l'affichage d'une partir du tunnel, on s'arrete quand cete zone est vide.
On a donc besoin:
_ d'une fonction qui calcul la zone du trou du tunnel apres l'affichage d'un plan donn
_ d'une fonction qui calcule l'intersection de deux zones
Pour leur simplicit on va choisir des rectangles
*/

/* calcul du rectangle projet
  les points suivant sont sur le cercle: P=O+M*sin(a)+N*cos(a)
	Px=ox+sa*mx+ca+nx Py=oy+sa*my+ca+ny Pz=oz+sa*mz+ca+nz
	ex=cx+fx*Px/Py
	ey=cy-fy*Pz/Py
	on cherche les bornes de ex et de ey
	ex=exmin ou exmax <=> dex/da=0
	ey=eymin ou eymax <=> dey/da=0
	dex/da=fx*(Py * dPx/da - Px * dPy/da) / Py = 0
	<=> Py * dPx/da - Px * dPy/da = 0
	(camx-sanx)(oy+samy+cany)-(ox+samx+canx)(camy-sany)=0
	 camxoy+casamxmy+camxny-sanxoy-sanxmy-casanxny
	-camyox-casamymx-camynx+sanyox+sanymx+casanxny   =0
	ca(mxoy-myox)+sa(nyox-nxoy)+(ca+sa)(mxny-mynx)   =0
	ca(mxoy-myox)+sa(nyox-nxoy)=mynx-mxny

	(oymx-oxmy)ca + (-oynx+oxny)sa + (mxny-mynx)ca + (mxny-mynx)sa =0
	(oymx-oxmy)ca + (-oynx+oxny)sa = mynx-mxny et ca+sa=1
	U*ca+V*sa=W et ca+sa=1

	Ux+Vy=W d=1/sqrt(U+V),u=U*d,v=V*d,w=W*d => ux+vy=w (u+v=1)
	distance entre la droite et l'origine:d=abs(w)
	point a cette distance:
	vx-uy=0 && ux+vy=w
	=> x=uw
	=> y=vw
	si d>1 -> erreur
	les solutions sont (x,y)+-(1-d)*(-v,u)
	x1=uw-sqrt(1-sqr(d))*v y1=vw+sqrt(1-sqr(d))*u
	x2=uw+sqrt(1-sqr(d))*v y2=vw-sqrt(1-sqr(d))*u
*/

